package ch.ethz.matsim.projects.astra_2018_002.analysis.flow;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.matsim.api.core.v01.Id;
import org.matsim.api.core.v01.events.LinkEnterEvent;
import org.matsim.api.core.v01.events.handler.LinkEnterEventHandler;
import org.matsim.api.core.v01.network.Link;
import org.matsim.vehicles.Vehicle;
import org.matsim.vehicles.Vehicles;

import ch.ethz.matsim.projects.astra_2018_002.flow_efficiency.FlowEfficiencyCalculator;
import ch.ethz.matsim.projects.astra_2018_002.flow_efficiency.FlowEfficiencyConfigurator;

public class FlowListener implements LinkEnterEventHandler {
	private final double startTime;
	private final double interval;

	private final int numberOfLinks;
	private final int numberOfBins;

	private final Item[][] timeItems;
	private final Item[] totalItems;

	private final Map<Id<Link>, Integer> linkMap = new HashMap<>();
	private final List<Link> links;
	private final Set<Id<Link>> linkIds;

	private final FlowEfficiencyCalculator flowEfficiencyCalculator;

	private final Vehicle CONVENTIONAL_VEHICLE;
	private final Vehicle PRIVATE_AV;
	private final Vehicle SHARED_AV;

	public FlowListener(Vehicles vehicles, Collection<Link> links, double startTime, double endTime, double interval,
			FlowEfficiencyCalculator flowEfficiencyCalculator) {
		this.startTime = startTime;
		this.interval = interval;

		this.numberOfBins = (int) Math.ceil((endTime - startTime) / interval);
		this.numberOfLinks = links.size();

		this.links = links.stream().collect(Collectors.toList());
		this.linkIds = links.stream().map(Link::getId).collect(Collectors.toSet());

		for (int index = 0; index < links.size(); index++) {
			linkMap.put(this.links.get(index).getId(), index);
		}

		this.CONVENTIONAL_VEHICLE = vehicles.getFactory().createVehicle(Id.createVehicleId("conventional"),
				vehicles.getVehicleTypes().get(FlowEfficiencyConfigurator.CONVENTIONAL_VEHICLE_TYPE_ID));
		this.PRIVATE_AV = vehicles.getFactory().createVehicle(Id.createVehicleId("private_av"),
				vehicles.getVehicleTypes().get(FlowEfficiencyConfigurator.PRIVATE_AV_VEHICLE_TYPE_ID));
		this.SHARED_AV = vehicles.getFactory().createVehicle(Id.createVehicleId("shared_av"),
				vehicles.getVehicleTypes().get(FlowEfficiencyConfigurator.SHARED_AV_VEHICLE_TYPE_ID));

		this.totalItems = new Item[numberOfLinks];
		this.timeItems = new Item[numberOfLinks][numberOfBins];
		this.flowEfficiencyCalculator = flowEfficiencyCalculator;

		for (int linkIndex = 0; linkIndex < numberOfLinks; linkIndex++) {
			totalItems[linkIndex] = new Item();

			for (int timeIndex = 0; timeIndex < numberOfBins; timeIndex++) {
				timeItems[linkIndex][timeIndex] = new Item();
			}
		}
	}

	public class Item {
		public int numberOfConventionalVehicles;
		public int numberOfPrivateAutomatedVehicles;
		public int numberOfSharedAutomatedVehicles;

		public double numberOfPCUs;
	}

	@Override
	public void handleEvent(LinkEnterEvent event) {
		if (linkIds.contains(event.getLinkId())) {
			int linkIndex = linkMap.get(event.getLinkId());
			Link link = links.get(linkIndex);

			Vehicle vehicle = CONVENTIONAL_VEHICLE;
			String plainVehicleId = event.getVehicleId().toString();

			if (plainVehicleId.startsWith("av_")) {
				vehicle = SHARED_AV;
			} else if (plainVehicleId.startsWith("prav_")) {
				vehicle = PRIVATE_AV;
			}

			Item totalItem = totalItems[linkIndex];
			totalItem.numberOfPCUs += flowEfficiencyCalculator.calculateFlowEfficiency(vehicle, link);

			if (vehicle == CONVENTIONAL_VEHICLE) {
				totalItem.numberOfConventionalVehicles++;
			} else if (vehicle == SHARED_AV) {
				totalItem.numberOfPrivateAutomatedVehicles++;
			} else {
				totalItem.numberOfSharedAutomatedVehicles++;
			}

			int timeIndex = (int) Math.floor((event.getTime() - startTime) / interval);

			if (timeIndex > -1 && timeIndex < numberOfBins) {
				Item timeItem = timeItems[linkIndex][timeIndex];
				timeItem.numberOfPCUs += flowEfficiencyCalculator.calculateFlowEfficiency(vehicle, link);

				if (vehicle == CONVENTIONAL_VEHICLE) {
					timeItem.numberOfConventionalVehicles++;
				} else if (vehicle == SHARED_AV) {
					timeItem.numberOfPrivateAutomatedVehicles++;
				} else {
					timeItem.numberOfSharedAutomatedVehicles++;
				}
			}
		}
	}

	public int getNumberOfLinks() {
		return numberOfLinks;
	}

	public int getNumberOfBins() {
		return numberOfBins;
	}

	public Item getTotalItem(Id<Link> linkId) {
		int linkIndex = linkMap.get(linkId);

		if (linkIndex == -1) {
			throw new IllegalStateException();
		}

		return totalItems[linkIndex];
	}

	public Item[] getTimeItems(Id<Link> linkId) {
		int linkIndex = linkMap.get(linkId);

		if (linkIndex == -1) {
			throw new IllegalStateException();
		}

		return timeItems[linkIndex];
	}
}
