package ch.ethz.matsim.projects.astra_2018_002.analysis.flow;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.matsim.api.core.v01.Scenario;
import org.matsim.api.core.v01.network.Link;
import org.matsim.core.api.experimental.events.EventsManager;
import org.matsim.core.config.Config;
import org.matsim.core.config.ConfigUtils;
import org.matsim.core.events.EventsUtils;
import org.matsim.core.events.MatsimEventsReader;
import org.matsim.core.network.io.MatsimNetworkReader;
import org.matsim.core.scenario.ScenarioUtils;
import org.matsim.core.utils.geometry.geotools.MGC;
import org.matsim.core.utils.gis.PolylineFeatureFactory;
import org.matsim.core.utils.gis.ShapeFileWriter;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import com.vividsolutions.jts.geom.Coordinate;

import ch.ethz.matsim.baseline_scenario.config.CommandLine;
import ch.ethz.matsim.baseline_scenario.config.CommandLine.ConfigurationException;
import ch.ethz.matsim.projects.astra_2018_002.analysis.flow.FlowListener.Item;
import ch.ethz.matsim.projects.astra_2018_002.flow_efficiency.FlowEfficiencyCalculator;
import ch.ethz.matsim.projects.astra_2018_002.flow_efficiency.FlowEfficiencyConfigurator;
import ch.ethz.matsim.projects.astra_2018_002.flow_efficiency.StaticFlowEfficiencyCalculator;

public class RunFlowAnalysis {
	static public void main(String[] args) throws ConfigurationException, IOException {
		CommandLine cmd = new CommandLine.Builder(args) //
				.requireOptions("network-path", "events-path", "output-path") //
				.build();

		Config config = ConfigUtils.createConfig();
		Scenario scenario = ScenarioUtils.createScenario(config);

		new MatsimNetworkReader(scenario.getNetwork()).readFile(cmd.getOptionStrict("network-path"));
		FlowEfficiencyConfigurator.defineVehicleTypes(scenario);

		FlowEfficiencyCalculator flowEfficiencyCalculator = new StaticFlowEfficiencyCalculator();

		double startTime = 5.0 * 3600.0;
		double endTime = 22.0 * 3600.0;
		double interval = 3600.0;

		List<Link> links = new LinkedList<>();
		Set<String> types = new HashSet<>(Arrays.asList("motorway", "trunk", "primary", "secondary"));

		for (Link link : scenario.getNetwork().getLinks().values()) {
			String osmType = (String) link.getAttributes().getAttribute("osm:way:highway");

			if (osmType != null && types.contains(osmType)) {
				links.add(link);
			}
		}

		FlowListener flowListener = new FlowListener(scenario.getVehicles(), links, startTime, endTime, interval,
				flowEfficiencyCalculator);

		EventsManager eventsManager = EventsUtils.createEventsManager();
		eventsManager.addHandler(flowListener);
		new MatsimEventsReader(eventsManager).readFile(cmd.getOptionStrict("events-path"));

		CoordinateReferenceSystem crs = MGC.getCRS("EPSG:2056");

		int morningPeakIndex = 8 - 5;
		int eveningPeakIndex = 17 - 5;

		Collection<SimpleFeature> features = new ArrayList<>(links.size());
		PolylineFeatureFactory featureFactory = new PolylineFeatureFactory.Builder() //
				.setCrs(crs) //
				.setName("links") //
				.addAttribute("link_id", String.class) //
				.addAttribute("day_cv", Integer.class) //
				.addAttribute("day_prav", Integer.class) //
				.addAttribute("day_sav", Integer.class) //
				.addAttribute("day_pcu", Double.class) //
				.addAttribute("am_cv", Integer.class) //
				.addAttribute("am_prav", Integer.class) //
				.addAttribute("am_sav", Integer.class) //
				.addAttribute("am_pcu", Double.class) //
				.addAttribute("pm_cv", Integer.class) //
				.addAttribute("pm_prav", Integer.class) //
				.addAttribute("pm_sav", Integer.class) //
				.addAttribute("pm_pcu", Double.class) //
				.addAttribute("capacity", Double.class) //
				.addAttribute("freespeed", Double.class) //
				.addAttribute("lanes", Double.class) //
				.addAttribute("osm_type", String.class) //
				.create();

		for (Link link : links) {
			Item totalItem = flowListener.getTotalItem(link.getId());
			Item morningItem = flowListener.getTimeItems(link.getId())[morningPeakIndex];
			Item eveningItem = flowListener.getTimeItems(link.getId())[eveningPeakIndex];

			Coordinate fromCoord = new Coordinate(link.getFromNode().getCoord().getX(),
					link.getFromNode().getCoord().getY());
			Coordinate toCoord = new Coordinate(link.getToNode().getCoord().getX(), link.getToNode().getCoord().getY());

			String osmType = (String) link.getAttributes().getAttribute("osm:way:highway");

			if (osmType == null) {
				osmType = "unknown";
			}

			SimpleFeature feature = featureFactory.createPolyline(new Coordinate[] { fromCoord, toCoord },
					new Object[] { link.getId().toString(), //
							totalItem.numberOfConventionalVehicles, totalItem.numberOfPrivateAutomatedVehicles,
							totalItem.numberOfSharedAutomatedVehicles, totalItem.numberOfPCUs, //
							morningItem.numberOfConventionalVehicles, morningItem.numberOfPrivateAutomatedVehicles,
							morningItem.numberOfSharedAutomatedVehicles, morningItem.numberOfPCUs, //
							eveningItem.numberOfConventionalVehicles, eveningItem.numberOfPrivateAutomatedVehicles,
							eveningItem.numberOfSharedAutomatedVehicles, eveningItem.numberOfPCUs, //
							link.getCapacity(), link.getFreespeed(), link.getNumberOfLanes(), osmType //
					}, null);
			features.add(feature);

			List<String> row = new LinkedList<>();

			row.add(link.getId().toString());
			row.add(String.valueOf(totalItem.numberOfConventionalVehicles));
			row.add(String.valueOf(totalItem.numberOfPrivateAutomatedVehicles));
			row.add(String.valueOf(totalItem.numberOfSharedAutomatedVehicles));
			row.add(String.valueOf(totalItem.numberOfPCUs));
		}

		ShapeFileWriter.writeGeometries(features, cmd.getOptionStrict("output-path"));
	}
}
