package ch.ethz.matsim.projects.astra_2018_002.long_distance;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.apache.log4j.Logger;
import org.matsim.api.core.v01.Id;
import org.matsim.api.core.v01.Scenario;
import org.matsim.api.core.v01.TransportMode;
import org.matsim.api.core.v01.network.Link;
import org.matsim.api.core.v01.network.Network;
import org.matsim.api.core.v01.population.Person;
import org.matsim.api.core.v01.population.PopulationFactory;
import org.matsim.api.core.v01.population.PopulationWriter;
import org.matsim.core.config.Config;
import org.matsim.core.config.ConfigUtils;
import org.matsim.core.network.NetworkUtils;
import org.matsim.core.network.algorithms.TransportModeNetworkFilter;
import org.matsim.core.network.io.MatsimNetworkReader;
import org.matsim.core.population.io.PopulationReader;
import org.matsim.core.router.DijkstraFactory;
import org.matsim.core.router.StageActivityTypes;
import org.matsim.core.router.StageActivityTypesImpl;
import org.matsim.core.router.util.LeastCostPathCalculatorFactory;
import org.matsim.core.scenario.ScenarioUtils;
import org.matsim.pt.PtConstants;

import ch.ethz.matsim.baseline_scenario.config.CommandLine;
import ch.ethz.matsim.baseline_scenario.config.CommandLine.ConfigurationException;

public class MakeLongDistanceScenario {
	private final static Logger logger = Logger.getLogger(MakeLongDistanceScenario.class);

	static public void main(String[] args) throws ConfigurationException, InterruptedException {
		CommandLine cmd = new CommandLine.Builder(args) //
				.requireOptions("network-path", "population-path", "output-path") //
				.allowOptions("threads", "batch-size", "osm-types") //
				.build();

		Config config = ConfigUtils.createConfig();
		Scenario scenario = ScenarioUtils.createScenario(config);

		new MatsimNetworkReader(scenario.getNetwork()).readFile(cmd.getOptionStrict("network-path"));
		new PopulationReader(scenario).readFile(cmd.getOptionStrict("population-path"));

		for (Link link : scenario.getNetwork().getLinks().values()) {
			link.setLength(Math.max(1.0, link.getLength()));
		}

		Network roadNetwork = NetworkUtils.createNetwork();
		new TransportModeNetworkFilter(scenario.getNetwork()).filter(roadNetwork,
				Collections.singleton(TransportMode.car));

		int numberOfThreads = cmd.getOption("threads").map(Integer::parseInt)
				.orElse(Runtime.getRuntime().availableProcessors());
		int batchSize = cmd.getOption("batch-size").map(Integer::parseInt).orElse(20);

		StageActivityTypes stageActivityTypes = new StageActivityTypesImpl(PtConstants.TRANSIT_ACTIVITY_TYPE);
		PopulationFactory populationFactory = scenario.getPopulation().getFactory();

		LeastCostPathCalculatorFactory routerFactory = new DijkstraFactory();

		LinkInteractionFinder interactionFinder = new LinkInteractionFinder(numberOfThreads, batchSize,
				stageActivityTypes, populationFactory, routerFactory);

		Set<String> osmTypes = new HashSet<>(Arrays.asList("motorway", "trunk"));

		if (cmd.hasOption("osm-types")) {
			osmTypes = new HashSet<>(Arrays.asList(cmd.getOptionStrict("osm-types").split(",")).stream()
					.map(String::trim).collect(Collectors.toList()));
		}

		final Set<String> finalOsmTypes = new HashSet<>(osmTypes);

		Function<Link, Boolean> highwayFilter = (link) -> {
			String attribute = (String) link.getAttributes().getAttribute("osm:way:highway");
			return attribute != null && finalOsmTypes.contains(attribute);
		};

		Collection<Id<Person>> interactionIds = interactionFinder.findInteractingPersons(scenario.getPopulation(),
				roadNetwork, highwayFilter);

		Set<Id<Person>> removalIds = new HashSet<>(scenario.getPopulation().getPersons().keySet());
		removalIds.removeAll(interactionIds);

		logger.info(String.format("Found %d/%d persons which do have highway interactions (removing %d persons)",
				interactionIds.size(), scenario.getPopulation().getPersons().size(), removalIds.size()));
		removalIds.forEach(scenario.getPopulation()::removePerson);

		new PopulationWriter(scenario.getPopulation()).write(cmd.getOptionStrict("output-path"));
	}
}
